/**
 * ----------------------------------------------------------------------------------
 *
 * Copyright (C) 2010 runlevel3 GmbH
 *
 * ----------------------------------------------------------------------------------
 *
 * LICENSE:
 *
 * This file is part of Rubidium-ClockDistribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ----------------------------------------------------------------------------------
 *
 * AVR Studio 4.15
 *
 * ----------------------------------------------------------------------------------
 *
 * @copyright 2010 runlevel3 GmbH
 * @license   http://www.gnu.org/licenses/gpl.txt GNU General Public License v3
 * @link      http://www.runlevel3.de runlevel3
 **/


#include <string.h>

#include <util/delay.h>
#include <util/crc16.h>
#ifdef WATCHDOG
#include <avr/wdt.h>
#endif

#include <1wire/1wire-error.h>
#include <1wire/1wire.h>
#include <1wire/1wire-manage.h>
#include <1wire/1wire-devices.h>




unsigned char ds18x20_init(unsigned char bus_nr, w1_dev *rom)
{
	unsigned char err, data;

	err = ds18x20_read_ps(bus_nr, rom, &data);
	if (err) return err;
	if (!data) rom->flags |= DEV_FLAGS_PARASITE_PWR;
	return 0;
}


unsigned char ds18x20_convert(unsigned char bus_nr, w1_dev *rom, unsigned char par_power)
{
	unsigned char err;

    if (!rom) {
		err = w1_command(bus_nr, DS18X20_CONVERT_T, NULL);
	} else {
		if (rom->flags & DEV_FLAGS_NOT_FOUND) return REMOVED_ERR;
		err = w1_command(bus_nr, DS18X20_CONVERT_T, rom->id);
	}
	if (rom) rom->err = err;
	if (err) return err;
	if (par_power) {
		err = w1_parasite_power(bus_nr);					//parasite power on
		if (err) {
			if (rom) rom->err = err;
			return err;
		}
	}
	return 0;
}


unsigned char ds18x20_read(unsigned char bus_nr, w1_dev *rom)
{
	unsigned char cnt, crc8=0;

    if (!rom) return PARAM_ERR;
	if (rom->flags & DEV_FLAGS_NOT_FOUND) return REMOVED_ERR;

	rom->err = w1_command(bus_nr, DS18X20_READ, rom->id);	//read command
	if (rom->err) return rom->err;
	for (cnt=0;cnt<9;cnt++) {
		rom->value[cnt] = w1_byte_rd(bus_nr);					//read bytes
		crc8 = _crc_ibutton_update(crc8, rom->value[cnt]);
	}
	if (crc8) {
		rom->err = CRC_ERR;
		return CRC_ERR;
	}
	return 0;
}


unsigned char ds18x20_write(unsigned char bus_nr, w1_dev *rom, unsigned char *data)
{
	unsigned char err;

	if (!rom) {
		err = w1_command(bus_nr, DS18X20_WRITE, NULL);			//write command
	} else {
		if (rom->flags & DEV_FLAGS_NOT_FOUND) return REMOVED_ERR;
		err = w1_command(bus_nr, DS18X20_WRITE, rom->id);
	}
	if (rom) rom->err = err;
	if (err) return err;
	w1_byte_wr(bus_nr, data[0]);								//write Th byte
	w1_byte_wr(bus_nr, data[1]);								//write Tl byte
	if (rom->id[0] == CHIP_DS18B20) w1_byte_wr(bus_nr, data[2]); //write Config byte
	return 0;
}


unsigned char ds18x20_read_ps(unsigned char bus_nr, w1_dev *rom, unsigned char *data)
{
	unsigned char err;

    if (!rom) {
		err = w1_command(bus_nr, DS18X20_READ_PS, NULL);
	} else {
		if (rom->flags & DEV_FLAGS_NOT_FOUND) return REMOVED_ERR;
		err = w1_command(bus_nr, DS18X20_READ_PS, rom->id);
	}
	if (rom) rom->err = err;
	if (err) return err;
	data[0] = w1_byte_rd(bus_nr);
	return 0;
}


float ds18x20_compute(w1_dev *rom)
{
    if (!rom) return PARAM_ERR;
	if (rom->id[0] == CHIP_DS18B20) return 0.0625 * (float)(signed int)(rom->value[0] | ((unsigned int)rom->value[1] << 8));
	//extended resolution
	if (rom->id[0] == CHIP_DS18S20) return (float)(signed char)((rom->value[0] >> 1) | (rom->value[1] << 7)) - 0.25 + 0.0625 * (float)(16 - rom->value[6]);
	return 0;
}
