/**
 * ----------------------------------------------------------------------------------
 *
 * Copyright (C) 2010 runlevel3 GmbH
 *
 * ----------------------------------------------------------------------------------
 *
 * LICENSE:
 *
 * This file is part of Rubidium-ClockDistribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ----------------------------------------------------------------------------------
 *
 * AVR Studio 4.15
 *
 * ----------------------------------------------------------------------------------
 *
 * @copyright 2010 runlevel3 GmbH
 * @license   http://www.gnu.org/licenses/gpl.txt GNU General Public License v3
 * @link      http://www.runlevel3.de runlevel3
 **/


#include <stdlib.h>

#include <avr/wdt.h>

#include <util/delay.h>

#include <1wire/1wire-config.h>
#include <1wire/1wire-error.h>
#include <1wire/1wire.h>

#ifdef W1_DISABLE_INT
#include <avr/interrupt.h>
#endif


unsigned char w1_bit_io(unsigned char bus_nr, unsigned char bit)
{
	unsigned char pinmask=1<<(W1_PIN+bus_nr);

	#ifdef W1_DISABLE_INT
	cli();
	#endif
	W1_OUT &= ~pinmask;						//output 0
	W1_DDR |= pinmask;
	_delay_us(3);							//3us start timeslot low pulse
	if (bit) W1_DDR &= ~pinmask;			//output bit

	_delay_us(7);							//sample after 3+7us
	if ((W1_IN & pinmask) == 0) {
		bit=0;								//sample bit
	} else bit = 1;

	#ifdef W1_DISABLE_INT
	sei();
	#endif
	_delay_us(50);							//remain output until 60us reached
	W1_DDR &= ~pinmask;						//disable output

	_delay_us(60);							//60us recovery time
	return bit;

}

unsigned char w1_reset(unsigned char bus_nr, unsigned char masterreset)
{
	unsigned char err, cnt, pinmask=1<<(W1_PIN+bus_nr);

	if (masterreset) {
		W1_OUT &= ~pinmask;
		W1_DDR |= pinmask;
		for (cnt=0;cnt<10;cnt++) {
			_delay_us(10000);				//power-on-reset for coupler and parasite powered devices
		}
		W1_OUT |= pinmask;					//reset release and strong pullup for 15us
		_delay_us(15);						//strong pullup delay
		W1_DDR &= ~pinmask;					//reset release
		W1_OUT &= ~pinmask;
		_delay_us(10000);					//wait for stable levels and double presence pulse
	}

	W1_OUT &= ~pinmask;
	W1_DDR |= pinmask;
	_delay_us(480);							//standard reset pulse
	W1_OUT |= pinmask;						//reset release and strong pullup for 15us
	_delay_us(15);							//strong pullup delay
	W1_DDR &= ~pinmask;						//reset release
	W1_OUT &= ~pinmask;
	_delay_us(60);							//presence detect delay
	err = W1_IN & pinmask;					//sample presence detect
	#ifdef WATCHDOG
		wdt_reset();
	#endif
	_delay_us(480-75);
	if (err) err = PRESENCE_ERR;
	if((W1_IN & pinmask)==0)
		err = SHORTC_ERR;					//short circuit detect
	return err;
}

unsigned char w1_parasite_power(unsigned char bus_nr)
{
	if (W1_IN & (1<<(W1_PIN+bus_nr))) {
		W1_OUT |= 1<<(W1_PIN+bus_nr);		//strong pullup
		W1_DDR |= 1<<(W1_PIN+bus_nr);		//parasite power on
		return 0;
	} else {
		return BUSY_ERR;
	}
}

void w1_parasite_power_off(unsigned char bus_nr)
{
	W1_DDR &= ~(1<<(W1_PIN+bus_nr));		//parasite power off
	W1_OUT &= ~(1<<(W1_PIN+bus_nr));
}

unsigned char w1_byte_wr(unsigned char bus_nr, unsigned char byte)
{
	unsigned char cnt, bit;

	for (cnt=0;cnt<8;cnt++) {
		bit = w1_bit_io(bus_nr, byte & 1);
		byte >>= 1;
		if (bit) byte |= 0x80;
	}
	return byte;
}

unsigned char w1_byte_rd(unsigned char bus_nr)
{
	return w1_byte_wr(bus_nr, 0xFF);
}

unsigned char w1_rom_search(unsigned char bus_nr, unsigned long long *branch, unsigned char *id)
{
	unsigned char i, j, bit=0, bits;
	unsigned long long mask=1;

	w1_byte_wr(bus_nr, SEARCH_ROM);			//ROM search command
	
	for (i=0;i<8;i++) {						//8 bytes
		for (j=0;j<8;j++) {					//8 bits
			bits = w1_bit_io(bus_nr, 1)<<1;	//read 2 bits
			bits |= w1_bit_io(bus_nr, 1);
			switch (bits) {
				case 0:
						if(*branch & mask) {
							bit=1;
						} else {
							bit=0;
						}
						break;
				case 1:
				case 2:
						*branch &= ~mask;
						bit=bits>>1;
						break;
				case 3:
						return DATA_ERR_SEARCH;	//1 1 = data error
						break;
			}
			*id >>= 1;
			w1_bit_io(bus_nr,bit);			//write bit
			if(bit) *id |= 0x80;			//store bit
			mask<<=1;
		}
		id++;								//next byte
	}
	for (mask=1ULL<<63;mask;mask>>=1) {		//search for right most 1, reset and set all bits right of this to 1
		if (*branch&mask) {
			*branch ^= mask;
			return 0;						//continue search
		} else {
			*branch |= mask;
		}
	}
	*branch=0;
	return 0;
}

unsigned char w1_command(unsigned char bus_nr, unsigned char command, unsigned char *id)
{
	unsigned char err, cnt;

	err = w1_reset(bus_nr, 0);
	if (err) return err;					//error

	if (id) {
		w1_byte_wr(bus_nr, MATCH_ROM);		//to a single device
		for (cnt=0;cnt<8;cnt++) {
			w1_byte_wr(bus_nr, *id);
			id++;
		}
	} else {
		w1_byte_wr(bus_nr, SKIP_ROM);		//to all devices
	}
	w1_byte_wr(bus_nr, command);
	return 0;
}
