/**
 * ----------------------------------------------------------------------------------
 *
 * Copyright (C) 2010 runlevel3 GmbH
 *
 * ----------------------------------------------------------------------------------
 *
 * LICENSE:
 *
 * This file is part of Rubidium-ClockDistribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ----------------------------------------------------------------------------------
 *
 * AVR Studio 4.15
 *
 * ----------------------------------------------------------------------------------
 *
 * @copyright 2010 runlevel3 GmbH
 * @license   http://www.gnu.org/licenses/gpl.txt GNU General Public License v3
 * @link      http://www.runlevel3.de runlevel3
 **/


#include <stdlib.h>
#include <math.h>

#include <avr/io.h>

#include <cy22150.h>
#include <twi_rtc.h>



pll_settings cy_pll;
unsigned char F2_enable = 1;
unsigned char cy22150_buf[256];


void cy22150_init(void)
{
    cy22150_buf[0x13] = CAPLOAD_DEFAULT;
    cy22150_buf[0x12] |= XCODRV_DEFAULT << 3;
	cy_pll.f_ref = f_reference;
}

unsigned long long computefrequency(unsigned char freq_type)
{
	unsigned int Q, PB, PO, P;
	unsigned long long FPLL;

    Q = (cy22150_buf[0x42] & 0x7F) + 2;
	if (freq_type == FTYPE_Q) return Q;
    PB = ((cy22150_buf[0x40] & 3) << 8) | cy22150_buf[0x41];
    PO = cy22150_buf[0x42] & 0x80 ? 1 : 0;
    P = (2 * (PB + 4)) + PO;
	if (freq_type == FTYPE_P) return P;
	FPLL = (P * f_reference_i * 1000ULL) / Q;
	
	if (freq_type == FTYPE_PLL) return FPLL;

	if (freq_type == FTYPE_F1) {
		if (cy22150_buf[0x0C] & 0x80) return (f_reference_i * 1000ULL) / (cy22150_buf[0x0C] & 0x7F);
		return FPLL / (cy22150_buf[0x0C] & 0x7F);
	}

	if (freq_type == FTYPE_F2) {
		if (cy22150_buf[0x47] & 0x80) return (f_reference_i * 1000ULL) / (cy22150_buf[0x47] & 0x7F);
		return FPLL / (cy22150_buf[0x47] & 0x7F);
	}

	return 0;
}


unsigned char checkPLL(void)
{
	unsigned long long p_tmp = computefrequency(FTYPE_P);
	unsigned long long pll_tmp = computefrequency(FTYPE_PLL);

	//check plausibility
	if ((p_tmp < 16ULL) || (p_tmp > 1024ULL) || (pll_tmp < 100000000000ULL) ||
		(pll_tmp > 400000000000ULL)) return 0;

	return 1;
}


void findPLLsettings(pll_settings *pll_vars)
{
    #define d_round     0.5
    #define qmin        2               // q range
    #define qmaxmax     129
    #define qmaxf       250000
    #define pmin        8               // p range
    #define pmax        2055
    #define dmin        2               // divider range
    #define dmax        128

	// constant values
	long f_pll_min =  25000000;			// 100 MHz, 25 MHz extended Range
	long f_pll_max = 450000000;			// 400 MHz, 450 MHz extended Range

    // precalculated constant values
    double f1f = pll_vars->f1_importance ? 10000.0 : 1;
    double f2f = pll_vars->f2_importance ? 10000.0 : 1;
    double pminf = f_pll_min / pll_vars->f_ref;
    double qminf = pll_vars->f_ref / f_pll_max;
    double qmax = pll_vars->f_ref / qmaxf;
    if (qmax > qmaxmax) qmax = qmaxmax;
    double f1bak = pll_vars->f1;
    double f2bak = pll_vars->f2;

    // start conditions
    double bestfrac = -1;
    double frac = -1;
    pll_vars->frac = -1;
    double q = qmin;
    double p = (int)(q * pminf);
    if (p < pmin) p = pmin ;
    int n = 0;

    if (!F2_enable) pll_vars->f2 = pll_vars->f1;

    if (pll_vars->f1 < f_pll_min / 127 && pll_vars->f2 > f_pll_min / 127) pll_vars->f1 = pll_vars->f2;
    if (pll_vars->f2 < f_pll_min / 127 && pll_vars->f1 > f_pll_min / 127) pll_vars->f2 = pll_vars->f1;

	DDRD |= _BV(PD6);

    while (1) {
		OCR2B++;

        n++;
        if (q > qmax) break;

        double d1 = (pll_vars->f_ref * p) / (pll_vars->f1 * q) + d_round;
        double d2 = (pll_vars->f_ref * p) / (pll_vars->f2 * q) + d_round;

        if (p > pmax || q < p * qminf || d1 >= dmax || d2 >= dmax) {
            q++;
            p = (int)(q * pminf);
            if (p < pmin) p = pmin;
            continue;
        }

        if (p < q * pminf || d1 < dmin || d2 < dmin || (int)(d2) == 3) {
            p++;
            continue;
        }

        double f1t = pll_vars->f_ref * p / (q * (int)(d1));
        double f2t = pll_vars->f_ref * p / (q * (int)(d2));
        if (F2_enable) frac = fabs(pll_vars->f1 - f1t) * f1f + fabs(pll_vars->f2 - f2t) * f2f;
        if (!F2_enable) frac = fabs(pll_vars->f1 - f1t) * f1f;
        if (bestfrac < 0 || frac < bestfrac) {
            bestfrac = frac;
            pll_vars->frac = frac;
            pll_vars->d1 = (int)(d1);
            if (F2_enable) pll_vars->d2 = (int)(d2);
            pll_vars->p = (int)(p);
            pll_vars->q = (int)(q);
            if (!frac) break;
        }
        p++;
    }
    pll_vars->iterations = n;

    pll_vars->f1 = f1bak;
    pll_vars->f2 = f2bak;

    if (pll_vars->f1 < f_pll_min / 127) {
        cy22150_buf[0x0C] |= 0x80;
        pll_vars->d1 = (int)(round(pll_vars->f_ref / pll_vars->f1));
        if (abs(pll_vars->d1) > 127) pll_vars->d1 = 127;
    } else {
        cy22150_buf[0x0C] &= 0x7F;
    }

    if (F2_enable) {
        if (pll_vars->f2 < f_pll_min / 127) {
            cy22150_buf[0x47] |= 0x80;
            pll_vars->d2 = (int)(round(pll_vars->f_ref / pll_vars->f2));
            if (abs(pll_vars->d2) > 127) pll_vars->d2 = 127;
        } else {
            cy22150_buf[0x47] &= 0x7F;
        }
    }
    setPQD1D2(pll_vars, 1);
    setChargePump(pll_vars);
}


void getOutput(unsigned char output, int *out_src)
{
	unsigned long out_bits = 0;

	if (!(cy22150_buf[0x09] & (1 << output))) {
		*out_src = 8;
		return;
	}

	out_bits = ((unsigned long)cy22150_buf[0x44] << 10) | ((unsigned long)cy22150_buf[0x45] << 2) | (((unsigned long)cy22150_buf[0x46] >> 6) & 3);
	*out_src = (out_bits & (7UL << (15 - (3 * output)))) >> (15 - (3 * output));
}


void setOutput(unsigned char output, int *out_src)
{
	unsigned long out_bits = 0;

	if (*out_src == 8) {
		cy22150_buf[0x09] &= ~(1 << output);
		return;
	} else {
		cy22150_buf[0x09] |= 1 << output;
	}

	out_bits = ((unsigned long)cy22150_buf[0x44] << 10) | ((unsigned long)cy22150_buf[0x45] << 2) | (((unsigned long)cy22150_buf[0x46] >> 6) & 3);
	out_bits &= ~(7UL << (15 - (3 * output)));
	out_bits |= (unsigned long)*out_src << (15 - (3 * output));

	cy22150_buf[0x44] = (out_bits >> 10) & 0xFF;
	cy22150_buf[0x45] = (out_bits >> 2) & 0xFF;
	cy22150_buf[0x46] &= ~(3 << 6);
	cy22150_buf[0x46] |= (out_bits & 3) << 6;
}


//extract p and q vars from buffer
void getPQD1D2(pll_settings *pll_vars)
{
	int PB, PO;

	//fill values in cy_pll structure
    cy_pll.q = (cy22150_buf[0x42] & 0x7F) + 2;

    PB = ((cy22150_buf[0x40] & 3) << 8) | cy22150_buf[0x41];
    PO = cy22150_buf[0x42] & 0x80 ? 1 : 0;
    cy_pll.p = (2 * (PB + 4)) + PO;

	cy_pll.d1 = cy22150_buf[0x0C] & 0x7F;
	cy_pll.d2 = cy22150_buf[0x47] & 0x7F;
}


//set buffer from computed p and q vars
void setPQD1D2(pll_settings *pll_vars, unsigned char set_out)
{
	if (set_out) {
		//set CLKOE bits LCLK1, LCLK2, CLK5, CLK6 enable
		cy22150_buf[0x09] |= 0x33;
	}

   	//reset divider
    cy22150_buf[0x0C] &= 0x80;
	if (set_out) {
	    //reset switch matrix
		cy22150_buf[0x44] &= 0x1F;
	    cy22150_buf[0x45] &= 0xF1;
    }

	if (pll_vars->d1 < 4) {
        if (pll_vars->d1 == 2) {
            cy22150_buf[0x0C] |= 4;
			if (set_out) {
				//set LCLK1 clksrc DIV1CLK/2
	            cy22150_buf[0x44] |= 0x40;
				//set  CLK5 clksrc DIV1CLK/2
	            cy22150_buf[0x45] |= 0x04;
			}
        }
        if (pll_vars->d1 == 3) {
            cy22150_buf[0x0C] |= 6;
			if (set_out) {
				//set LCLK1 clksrc DIV1CLK/3
	            cy22150_buf[0x44] |= 0x60;
				//set  CLK5 clksrc DIV1CLK/3
	            cy22150_buf[0x45] |= 0x06;
			}
        }
    } else {
        cy22150_buf[0x0C] |= pll_vars->d1 & 0x7F;
		if (set_out) {
			//set LCLK1 clksrc DIV1CLK/DIV1N
	        cy22150_buf[0x44] |= 0x20;
			//set  CLK5 clksrc DIV1CLK/DIV1N
	        cy22150_buf[0x45] |= 0x02;
		}
    }

    if (F2_enable) {
    	//reset divider
	    cy22150_buf[0x47] &= 0x80;
		if (set_out) {
		    //reset switch matrix
		    cy22150_buf[0x44] &= 0xE3;
		    cy22150_buf[0x45] &= 0xFE;
		    cy22150_buf[0x46] &= 0x3F;
		}
        if (pll_vars->d2 < 5) {
            if (pll_vars->d2 == 2) {
                cy22150_buf[0x47] |= 4;
				if (set_out) {
					//set LCLK2 clksrc DIV2CLK/2
	    	        cy22150_buf[0x44] |= 0x14;
					//set  CLK6 clksrc DIV2CLK/2
	    	        cy22150_buf[0x45] |= 0x01;
	    	        cy22150_buf[0x46] |= 0x40;
				}
            }
            if (pll_vars->d2 == 4) {
                cy22150_buf[0x47] |= 8;
				if (set_out) {
					//set LCLK2 clksrc DIV2CLK/4
	    	        cy22150_buf[0x44] |= 0x18;
					//set  CLK6 clksrc DIV2CLK/4
	    	        cy22150_buf[0x45] |= 0x01;
	    	        cy22150_buf[0x46] |= 0x80;
				}
            }
        } else {
            cy22150_buf[0x47] |= pll_vars->d2 & 0x7F;
			if (set_out) {
				//set LCLK2 clksrc DIV2CLK/DIV2N
	            cy22150_buf[0x44] |= 0x10;
				//set  CLK6 clksrc DIV2CLK/DIV2N
	   	        cy22150_buf[0x45] |= 0x01;
			}
        }
    }

    if (pll_vars->frac >= 0) {
        int pb;
        // (2*(PB+4)+PO) * Fref / (Q+2)
        pb = (pll_vars->p >> 1) - 4;
        if (pll_vars->p & 1) {
            cy22150_buf[0x42] |= 0x80;
        } else {
            cy22150_buf[0x42] &= 0x7F;
        }
        cy22150_buf[0x40] &= 0xFC;
        cy22150_buf[0x40] |= (pb >> 8) & 3;
        cy22150_buf[0x41] = pb & 0xFF;
        cy22150_buf[0x42] &= 0x80;
        cy22150_buf[0x42] |= (pll_vars->q - 2) & 0x7F;
    }
}

void setChargePump(pll_settings *pll_vars)
{
    cy22150_buf[0x40] &= 0xE3;
    switch (pll_vars->p) {
        case 0 ... 15:
			//unstable charge-pump-setting, p < 15
            break;
        case 16 ... 44:
            break;
        case 45 ... 479:
            cy22150_buf[0x40] |= 1 << 2;
            break;
        case 480 ... 639:
            cy22150_buf[0x40] |= 2 << 2;
            break;
        case 640 ... 799:
            cy22150_buf[0x40] |= 3 << 2;
            break;
        case 800 ... 1023:
            cy22150_buf[0x40] |= 4 << 2;
            break;
        case 1024 ... 1319:
			//unstable charge-pump-setting, p > 1023
            cy22150_buf[0x40] |= 5 << 2;
            break;
        case 1320 ... 1679:
			//unstable charge-pump-setting, p > 1320
            cy22150_buf[0x40] |= 6 << 2;
            break;
        case 1680 ... 2055:
			//unstable charge-pump-setting, p > 1680
            cy22150_buf[0x40] |= 7 << 2;
            break;
        default:
			//invalid charge-pump-setting, p out of range
            break;
    }

}

void cy22150_buf_write(void)
{
	rtc_write_bytes(TWI_SLA_CY22150, 0x0C, 1, &cy22150_buf[0x0C]);
	rtc_write_bytes(TWI_SLA_CY22150, 0x12, 2, &cy22150_buf[0x12]);
	rtc_write_bytes(TWI_SLA_CY22150, 0x40, 3, &cy22150_buf[0x40]);
	rtc_write_bytes(TWI_SLA_CY22150, 0x44, 3, &cy22150_buf[0x44]);
	rtc_write_bytes(TWI_SLA_CY22150, 0x47, 1, &cy22150_buf[0x47]);
	rtc_write_bytes(TWI_SLA_CY22150, 0x09, 1, &cy22150_buf[0x09]);
}
